//------------------------------------------------------------------------------------------------
//
//   Project Blue - Pneumatic Tube Wire Renderer
//
//------------------------------------------------------------------------------------------------

package gcewing.projectblue;

import net.minecraft.client.renderer.texture.IIconRegister;
import gcewing.projectblue.PneumaticTubePart.InsulatedTubeWire;

public class PneumaticTubeWireRenderer {

	/* Connection Mask */

	/**
	 * Split into 4 nybbles: (from lowest)
	 * 0 = Corner connections (this wire should connect around a corner to something external)
	 * 1 = Straight connections (this wire should connect to something external)
	 * 2 = Internal connections (this wire should connect to something internal)
	 * 3 = Internal open connections (this wire is not blocked by a cover/edge part and *could* connect through side)
	 * bit 16 = connection to the centerpart
	 * 5 = Render corner connections. Like corner connections but set to low if the other wire part is smaller than this (they render to us not us to them)
	 */
	 
	 final static int CORNER_BIT        = 0x1;
	 final static int STRAIGHT_BIT      = 0x10;
	 final static int INTERNAL_BIT      = 0x100;
	 final static int OPEN_BIT          = 0x1000;
	 final static int CENTER_BIT        = 0x10000;
	 final static int RENDER_CORNER_BIT = 0x100000;
	 
	 final static int STUB_BIT          = 0x1000000; // Used internally by this renderer

	static PBTexture wireTexture, endTextures[];
	static PBModel[]
		fillModels, 
		stubModels, 
		straightModels, 
		cornerModels, 
		internalModels;
	
	public static void init() {
		fillModels = getModels("tubewire_fill_nw", "tubewire_fill_se");
		stubModels = getModels("tubewire_stub_ne", "tubewire_stub_sw");
		straightModels = getModels("tubewire_straight_ne", "tubewire_straight_sw");
		cornerModels = getModels("tubewire_corner_ne", "tubewire_corner_sw");
		internalModels = getModels("tubewire_internal_ne", "tubewire_internal_sw");
	}
	
	static PBModel[] getModels(String... names) {
		PBModel[] result = new PBModel[names.length];
		for (int i = 0; i < names.length; i++)
			result[i] = ProjectBlue.getModel(names[i]);
		return result;
	}

	public static void registerIcons(IIconRegister reg) {
//		wireTexture = getTexture(reg, "tubewire");
		endTextures = new PBTexture[] {
			getTexture(reg, "redwire_end_off"),
			getTexture(reg, "redwire_end_on")
		};
	}

	static PBTexture getTexture(IIconRegister reg, String name) {
		return new PBTexture(ProjectBlue.mod.getIcon(reg, name));
	}
	
	static int[][] rotToTurn = {
		{3, 2, 1, 0}, // DOWN
		{1, 0, 3, 2}, // UP
		{1, 0, 3, 2}, // NORTH
		{1, 0, 3, 2}, // SOUTH
		{1, 0, 3, 2}, // WEST
		{1, 0, 3, 2}, // EAST
	};
	
	public static void render(InsulatedTubeWire wire, double x, double y, double z, IPBRenderer pbr) {
		int side = wire.side();
		int rconn = wire.connMap();
		if ((rconn & 0xfff) == 0)
			// Render as a pair of stubs
			rconn = 0x5 * STUB_BIT;
		else {
			// If there is only one connection, render a stub opposite it
			int rstub = -1;
			for (int r = 0; r < 4; r++) {
				if (((rconn >> r) & 0x111) != 0) {
					if (rstub == -1)
						rstub = r ^ 2;
					else
						rstub = -2;
				}
			}
			if (rstub >= 0)
				rconn |= (STUB_BIT << rstub);
		}
		System.out.printf("PneumaticTubeWireRenderer.render: side = %s, rconn = 0x%x\n", side, rconn);
		double h = 0.501;
		Trans3 t = new Trans3(x + h, y + h, z + h).side(side).translate(0, -0.5, 0);
		for (int r = 0; r < 4; r++) {
			int sconn = rconn >> r;
			int turn = rotToTurn[side][r];
			int q = (turn >> 1) & 1;
			Trans3 ts = t.turn(turn);
			PBModel model;
			if ((sconn & STUB_BIT) != 0)
				model = stubModels[q];
			else if ((sconn & CORNER_BIT) != 0) {
				if ((sconn & RENDER_CORNER_BIT) != 0)
					model = cornerModels[q];
				else
					model = straightModels[q];
			}
			else if ((sconn & STRAIGHT_BIT) != 0)
				model = straightModels[q];
			else if ((sconn & INTERNAL_BIT) != 0)
				model = internalModels[q];
			else
				model = fillModels[((turn - 1) >> 1) & 1];
			PBTexture wireTexture = new PBTexture(wire.getWireType().wireSprites()[0]);
			int state = wire.signal() != 0 ? 1 : 0;
			PBTexture endTexture = endTextures[state];
			model.render(ts, pbr, wireTexture, endTexture);
		}
	}

}
